# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-12-06 23:39
from __future__ import unicode_literals

from django.db import migrations, models
from djeddit.models import Thread


def genSlugs(apps, schema_editor):
    for t in Thread.objects.all().iterator():
        t.save()


class Migration(migrations.Migration):
    dependencies = [
        ('djeddit', '0005_thread_is_stickied'),
    ]

    operations = [
        migrations.AddField(
            model_name='thread',
            name='slug',
            field=models.SlugField(null=True),
        ),
        migrations.RunPython(genSlugs)
    ]



from django.test import TestCase
from django import VERSION as DJANGO_VERSION

if DJANGO_VERSION[:2] < (1, 10):
    from django.core.urlresolvers import reverse
else:
    from django.urls import reverse
from djeddit.utils.base_tests import TestCalls, createUser
from djeddit.models import Topic, Thread, Post, UserPostVote
from djeddit.utils.utility_funcs import gen_uuid


# Create your tests here.


class CreateThreadTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/create_thread.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.topic = Topic.objects.create(title='Test Topic')
        cls.url = reverse('createThread', args=[cls.topic.urlTitle])

    def testLoads(self):
        self._test_call_view_loads(self.url)

    def _testSubmit(self, data):
        self._test_call_view_submit(self.url, code=302, data=data)
        thread = Thread.objects.get(topic=self.topic, title=data['thread-title'])
        self.assertEqual(thread.op.content, data['post-content'])
        self.assertEqual(thread.op.ip_address, '127.0.0.1')
        self.assertEqual(thread.op.user_agent, None)

    def testSubmitUnauthenticated(self):
        data = {'post-content': 'content', 'thread-title': 'Thread1', 'thread-url': ''}
        self._testSubmit(data)

    def testSubmitAuthenticated(self):
        data = {'post-content': 'content', 'thread-title': 'Thread1', 'thread-url': ''}
        self.login()
        self._testSubmit(data)


class DeleteTopicTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self)

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='admin', email='admin@example.com', password='pass', is_superuser=True)
        cls.topic = Topic.objects.create(title='Test Topic')
        cls.url = reverse('deleteTopic', args=[cls.topic.urlTitle])

    def testDeleteTopic(self):
        self.login()
        self._test_call_view_redirects(self.url, reverse('topics'))
        self.assertRaises(Topic.DoesNotExist, self.topic.refresh_from_db)

    def testUnknownTopic(self):
        self.login()
        url = reverse('deleteTopic', args=['Fake_Topic'])
        self._test_call_view_code(url, 404, post=True)

    def testRequireLogin(self):
        self._test_call_view_redirected_login(self.url)

    def testRequireSuperUser(self):
        self._create_user_and_login()
        self.login('user', 'pass')
        self._test_call_view_redirected_login(self.url)


class LockThreadTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self)

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='admin', email='admin@example.com', password='pass', is_superuser=True)
        topic = Topic.objects.create(title='Test Topic')
        cls.thread = Thread.objects.create(title='Test_Thread', topic=topic, op=Post.objects.create())
        cls.url = reverse('lockThread', args=[cls.thread.id])

    def _testLockAndUnlock(self):
        self.login()
        # lock thread
        self._test_call_view_code(self.url, 302)
        # test thread is locked
        replyUrl = reverse('replyPost', args=[self.thread.op.uid])
        self._test_call_view_code(replyUrl, 403, data=dict(content='replied'), post=True)
        editUrl = reverse('editPost', args=[self.thread.op.uid])
        self._test_call_view_code(editUrl, 403, data={'post-content': 'edited'}, post=True)
        # unlock thread
        self._test_call_view_code(self.url, 302)
        # test thread is unlocked
        self._test_call_view_code(replyUrl, 302, data=dict(content='replied'), post=True)
        Post.objects.get(parent=self.thread.op, content='replied')  # assert object exists
        self._test_call_view_code(editUrl, 302, data={'post-content': 'edited'}, post=True)
        self.thread.op.refresh_from_db()
        self.assertEqual(self.thread.op.content, 'edited')

    def _testRequireSuperUser(self):
        self._setup_user('not_admin', 'not_admin@example.com', password='pass')
        self.login()
        self._test_call_view_redirected_login(self.url)

    def runTestsInSequence(self):
        self._testLockAndUnlock()
        self._testRequireSuperUser()

    def testRequireLogin(self):
        self._test_call_view_redirected_login(self.url)


class StickyThreadTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self)

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='admin', email='admin@example.com', password='pass', is_superuser=True)
        topic = Topic.objects.create(title='Test Topic')
        cls.thread = Thread.objects.create(title='Test_Thread', topic=topic, op=Post.objects.create())
        cls.url = reverse('stickyThread', args=[cls.thread.id])

    def _testStickyAndUnsticky(self):
        self.login()
        self.assertEqual(self.thread.is_stickied, False)
        self._test_call_view_code(self.url, 302) # set sticky to true
        self.assertEqual(self.thread.is_stickied, True)
        self._test_call_view_code(self.url, 302)  # set sticky to false
        self.assertEqual(self.thread.is_stickied, False)

    def _testRequireSuperUser(self):
        self._setup_user('not_admin', 'not_admin@example.com', password='pass')
        self.login()
        self._test_call_view_redirected_login(self.url)

    def runTestsInSequence(self):
        self._testStickyAndUnsticky()
        self._testRequireSuperUser()

    def testRequireLogin(self):
        self._test_call_view_redirected_login(self.url)


class TopicsPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/topics.html')

    def testLoads(self):
        url = reverse('topics')
        self._test_call_view_loads(url)


class TopicPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/topic.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='admin', email='admin@example.com', password='pass', is_superuser=True)
        cls.topic = Topic.objects.create(title='Test Topic')
        cls.url = reverse('topicPage', args=[cls.topic.urlTitle])

    def testLoads(self):
        self._test_call_view_loads(self.url)

    def testRedirectsOldTitle(self):
        url = reverse('topicPage', args=[self.topic.title.replace(' ', '_')])
        self._test_call_view_redirects(url, redirected_url=self.url)

    def testUnknownTopic(self):
        url = reverse('topicPage', args=['Fake_Topic'])
        self._test_call_view_code(url, 404)

    def testEditTopic(self):
        self.login()
        data = dict(title='Test Topic edited', description='Test Description')
        self._test_call_view_submit(self.url, code=302, data=data)
        self.topic.refresh_from_db()
        self.assertEqual(self.topic.title, data['title'])
        self.assertEqual(self.topic.description, data['description'])

    def testEditTopicRequireLogin(self):
        data = dict(title='Test Topic (edited)', description='Test Description')
        self._test_call_view_submit(self.url, code=403, data=data)
        self.topic.refresh_from_db()
        self.assertNotEqual(self.topic.title, data['title'])
        self.assertNotEqual(self.topic.description, data['description'])

    def testEditTopicRequireSuperUser(self):
        self._create_user_and_login()
        self.login('user', 'pass')
        self.testEditTopicRequireLogin()


class ThreadPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/thread.html')

    @classmethod
    def setUpTestData(cls):
        cls.topic = Topic.objects.create(title='Test Topic')
        cls.thread = Thread.objects.create(title='Test_Thread', topic=cls.topic, op=Post.objects.create())

    def testLoads(self):
        self._test_call_view_loads(self.thread.relativeUrl)

    def testWrongSlug(self):
        # test no slug
        url = reverse('threadPage', args=[self.topic.urlTitle, self.thread.pk])
        self._test_call_view_redirects(url, redirected_url=self.thread.relativeUrl)
        # test wrong slug
        url = reverse('threadPage', args=[self.topic.urlTitle, self.thread.pk, 'wrong-page'])
        self._test_call_view_redirects(url, redirected_url=self.thread.relativeUrl)

    def testWrongTopic(self):
        url = reverse('threadPage', args=['Fake_Topic', self.thread.id, 'fake-topic'])
        self._test_call_view_code(url, 404)

    def testOldTitleRedirect(self):
        url = reverse('threadPage', args=[self.topic.title.replace(' ', '_'), self.topic.pk, self.thread.slug])
        self._test_call_view_redirects(url, redirected_url=self.thread.relativeUrl)

    def testWrongThread(self):
        url = reverse('threadPage', args=[self.topic.urlTitle, self.thread.id + 1, 'fake-topic'])
        self._test_call_view_code(url, 404)


class ReplyPostTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/reply_form.html')

    @classmethod
    def setUpTestData(cls):
        cls.post = Post.objects.create()
        topic = Topic.objects.create(title='Test Topic')
        Thread.objects.create(title='Test_Thread', topic=topic, op=cls.post)
        cls.url = reverse('replyPost', args=[cls.post.uid])

    def testLoads(self):
        self._test_call_view_loads(self.url)

    def testSubmit(self):
        self._test_call_view_submit(self.url, code=302, data=dict(content='replied'))
        rp = Post.objects.get(parent=self.post, content='replied')  # assert object exists
        self.assertEqual(rp.ip_address, '127.0.0.1')
        self.assertEqual(rp.user_agent, None)

    def testUnknownUid(self):
        uid = gen_uuid()
        url = reverse('replyPost', args=[str(uid)])
        self._test_call_view_code(url, 404)

    def testNoThread(self):
        post = Post.objects.create()
        url = reverse('replyPost', args=[str(post.uid)])
        self._test_call_view_code(url, 404)


class EditPostTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/edit_post.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.post = Post.objects.create(created_by=cls.user)
        topic = Topic.objects.create(title='Test Topic')
        Thread.objects.create(title='Test_Thread', topic=topic, op=cls.post)
        user1 = createUser('user1', 'user1@example.com', 'pass')
        cls.User1Post = Post.objects.create(created_by=user1, parent=cls.post)
        cls.url = reverse('editPost', args=[cls.post.uid])

    def testLoads(self):
        self.login()
        self._test_call_view_loads(self.url)

    def testSubmit(self):
        self.login()
        self._test_call_view_submit(self.url, code=302, data={'post-content': 'edited'})
        self.post.refresh_from_db()
        self.assertEqual(self.post.content, 'edited')

    def testUnknownUid(self):
        uid = gen_uuid()
        self.login()
        url = reverse('editPost', args=[str(uid)])
        self._test_call_view_code(url, 404)

    def testNoThread(self):
        post = Post.objects.create()
        url = reverse('editPost', args=[str(post.uid)])
        self.login()
        self._test_call_view_code(url, 404)

    def testEditAnotherPost(self):
        url = reverse('editPost', args=[str(self.User1Post.uid)])
        self.login()
        self._test_call_view_code(url, 403)

    def testEditAnotherPostAsSuperuser(self):
        self._create_user_and_login(create_admin=True)
        url = reverse('editPost', args=[str(self.User1Post.uid)])
        self._test_call_view_submit(url, code=302, data={'post-content': 'edited'})
        self.User1Post.refresh_from_db()
        self.assertEqual(self.User1Post.content, 'edited')

    def testEditThread(self):
        self._create_user_and_login(create_admin=True)
        self._test_call_view_submit(self.url, code=302, data={'post-content': 'edited'})
        self.post.refresh_from_db()
        self.assertEqual(self.post.content, 'edited')


class VotePostTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self)

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.post = Post.objects.create()
        cls.url = reverse('votePost')

    def _testVote(self, submit_val, expected_score, expected_votePost_val, user=None, post=None):
        user = user or VotePostTest.user
        post = post or self.post
        self._test_call_view_submit(self.url, data=dict(post=post.uid, vote=submit_val))
        user_vote = UserPostVote.objects.get(user=user, post=post)  # assert object exists
        self.assertEqual(user_vote.val, expected_votePost_val)
        post.refresh_from_db()
        self.assertEqual(post.score, expected_score)

    def testSubmit(self):
        self.login()
        self.assertEqual(self.post.score, 0)
        self._testVote(1, 1, 1)  # test upvote
        self._testVote(2, 1, 1)  # test vote value above 1
        self._testVote(-1, -1, -1)  # test downvote
        self._testVote(-2, -1, -1)  # test downvote value below -1

    def testRequiresLogin(self):
        self._test_call_view_redirected_login(self.url, dict(post=self.post.uid, vote=1))

    def testEmptyData(self):
        self.login()
        self._test_call_view_code(self.url, 400, post=True)

    def testVoteOwnPost(self):
        self.login()
        post = Post.objects.create(created_by=VotePostTest.user)
        self._test_call_view_submit(self.url, code=403, data=dict(post=post.uid, vote=1))

    def testVoteOwnPostAsSuperuser(self):
        admin = self._create_user_and_login(create_admin=True)
        post = Post.objects.create(created_by=admin)
        self._testVote(1, 1, 1, user=admin, post=post)


class DeletePostTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self)

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='admin', email='admin@example.com', password='pass', is_superuser=True)
        topic = Topic.objects.create(title='Test Topic')
        cls.thread = Thread.objects.create(title='Test_Thread', topic=topic, op=Post.objects.create())
        cls.comment = Post.objects.create(parent=cls.thread.op)
        cls.url = reverse('deletePost', args=[cls.comment.uid])

    def testRequireLogin(self):
        self._test_call_view_redirected_login(self.url)

    def testRequireSuperUser(self):
        self._create_user_and_login()
        self._test_call_view_redirected_login(self.url)

    def testDeleteComment(self):
        self.login()
        self._test_call_view_redirects(self.url, self.thread.relativeUrl)
        self.assertRaises(Post.DoesNotExist, self.comment.refresh_from_db)

    def testDeleteThread(self):
        self.login()
        url = reverse('deletePost', args=[self.thread.op.uid])
        redirect_url = reverse('topicPage', args=[self.thread.topic.urlTitle])
        self._test_call_view_redirects(url, redirect_url)
        self.assertRaises(Post.DoesNotExist, self.thread.op.refresh_from_db)
        self.assertRaises(Thread.DoesNotExist, self.thread.refresh_from_db)


class UserSummaryTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/user_summary.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.topic = Topic.objects.create(title='Test Topic')
        cls.url = reverse('userSummary', args=[cls.username])

    def testLoadsEmpty(self):
        self._test_call_view_loads(self.url)

    def testLoads(self):
        op = Post.objects.create(created_by=self.user)
        Thread.objects.create(topic=self.topic, title='Test Thread', op=op)
        Post.objects.create(created_by=self.user, parent=op)
        self._test_call_view_loads(self.url)

    def testInvalidUser(self):
        url = reverse('userSummary', args=['invaliduser'])
        self._test_call_view_code(url, 404)


class UserThreadsPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/user_threads.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        cls.topic = Topic.objects.create(title='Test Topic')
        cls.url = reverse('userThreads', args=[cls.username])

    def testLoadsEmpty(self):
        self._test_call_view_loads(self.url)

    def testLoads(self):
        op = Post.objects.create(created_by=self.user, content='op content')
        Thread.objects.create(topic=self.topic, title='Test Thread', op=op)
        self._test_call_view_loads(self.url)

    def testInvalidUser(self):
        url = reverse('userSummary', args=['invaliduser'])
        self._test_call_view_code(url, 404)


class UserRepliesPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/user_threads.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='user', email='user@example.com', password='pass')
        topic = Topic.objects.create(title='Test Topic')
        cls.thread = Thread.objects.create(topic=topic, title='Test Thread', op=Post.objects.create())
        cls.url = reverse('userThreads', args=[cls.username])

    def testLoadsEmpty(self):
        self._test_call_view_loads(self.url)

    def testLoads(self):
        Post.objects.create(parent=self.thread.op)
        self._test_call_view_loads(self.url)

    def testInvalidUser(self):
        url = reverse('userSummary', args=['invaliduser'])
        self._test_call_view_code(url, 404)


class UsersPageTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self, 'djeddit/users_page.html')

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='admin', email='admin@example.com', password='pass', is_superuser=True)
        cls.url = reverse('usersPage')

    def testLoads(self):
        user = createUser()
        topic = Topic.objects.create(title='Test Topic')
        op = Post.objects.create(created_by=user)
        Thread.objects.create(topic=topic, title='Test Thread', op=op)
        Post.objects.create(created_by=user, parent=op)
        self.login()
        self._test_call_view_loads(self.url)

    def testRequireLogin(self):
        self._test_call_view_redirected_login(self.url)

    def testRequireSuperUser(self):
        self._create_user_and_login()
        self._test_call_view_redirected_login(self.url)


class SetUserStatusTest(TestCase, TestCalls):
    def __init__(self, *args, **kwargs):
        TestCase.__init__(self, *args, **kwargs)
        TestCalls.__init__(self)

    @classmethod
    def setUpTestData(cls):
        cls._setup_user(username='admin', email='admin@example.com', password='pass', is_superuser=True)
        cls.url = reverse('setUserStatus')

    def _testChangeStatus(self, user, status):
        self._test_call_view_submit(self.url, data=dict(username=user.username, status=status))
        user.refresh_from_db()
        if status == 'active':
            self.assertTrue(user.is_active)
            self.assertTrue(not user.is_superuser)
        elif status == 'banned':
            self.assertTrue(not user.is_active)
            self.assertTrue(not user.is_superuser)
        elif status == 'admin':
            self.assertTrue(user.is_active)
            self.assertTrue(user.is_superuser)
        else:
            raise ValueError('Invalid status: %s' % status)

    def testPromoteToAdmin(self):
        user = self._create_user_and_login()
        self.login()
        self._testChangeStatus(user, 'admin')
        self._testChangeStatus(user, 'active')
        self._testChangeStatus(user, 'banned')

    def testRequireLogin(self):
        self._test_call_view_redirected_login(self.url)

    def testRequireSuperUser(self):
        self._create_user_and_login()
        self._test_call_view_redirected_login(self.url)



from crispy_forms.layout import Submit
from crispy_forms.helper import FormHelper
from django import forms

from .models import Topic, Thread, Post


class TopicForm(forms.ModelForm):
    class Meta:
        model = Topic
        fields = ['title', 'description']

    def __init__(self, *args, **kwargs):
        super(TopicForm, self).__init__(*args, **kwargs)
        self.helper = FormHelper(self)
        self.helper.layout.append(Submit('submit', 'Submit', css_class='btn btn-primary'))


class ThreadForm(forms.ModelForm):
    class Meta:
        model = Thread
        fields = ['title', 'url']


class PostForm(forms.ModelForm):
    class Meta:
        model = Post
        fields = ['content']



from django import VERSION as DJANGO_VERSION

import uuid
from math import sqrt


def gen_uuid():
    return uuid.uuid4()


def wsi_confidence(ups, downs):
    """Implementation of Wilson score interval used for scoring posts"""
    n = ups + downs
    if n:
        z = 1.96 # 1.44 = 85%, 1.96 = 95%
        phat = float(ups) / n
        return (phat + z * z / (2 * n) - z * sqrt((phat * (1 - phat) + z * z / (4 * n)) / n)) / (1 + z * z / n)
    return 0


def is_authenticated(request):
    """user.is_authenticated check with added compatibility for older versions"""
    if DJANGO_VERSION[:2] < (1, 10):
        return request.user.is_authenticated()
    else:
        return request.user.is_authenticated



"""django1 URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.9/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
from django.conf.urls import url
from . import views
from django.conf import settings

if not hasattr(settings, 'TOPICS_URL'):
    settings.TOPICS_URL = ''
topic_prefix = '%s/' % settings.TOPICS_URL if settings.TOPICS_URL else ''


urlpatterns = [
    url(r'^%s$' % (topic_prefix + '/?' if topic_prefix else ''), views.topicsPage, name='topics'),
    url(r'^lock_thread/(\d+)?/?$', views.lockThread, name='lockThread'),
    url(r'^sticky_thread/(\d+)?/?$', views.stickyThread, name='stickyThread'),
    url(r'^reply_post/([\w\-]{36})?/?$', views.replyPost, name='replyPost'),
    url(r'^edit_post/([\w\-]{36})?/?$', views.editPost, name='editPost'),
    url(r'^vote_post/?$', views.votePost, name='votePost'),
    url(r'^delete_post/([\w\-]{36})?/?$', views.deletePost, name='deletePost'),
    url(r'^load_additional_replies/?$', views.loadAdditionalReplies, name='loadAdditionalReplies'),
    url(r'^user/(.+)/summary/?$', views.userSummary, name='userSummary'),
    url(r'^user/(.+)/threads/?$', views.userThreadsPage, name='userThreads'),
    url(r'^user/(.+)/replies/?$', views.userRepliesPage, name='userReplies'),
    url(r'^users/?$', views.usersPage, name='usersPage'),
    url(r'^set_user_status/?$', views.setUserStatus, name='setUserStatus'),
    url(r'^%s([a-zA-Z0-9\-_]+)/?$' % topic_prefix, views.topicPage, name='topicPage'),
    url(r'^%s([a-zA-Z0-9\-_]+)/delete_topic/?$' % topic_prefix, views.deleteTopic, name='deleteTopic'),
    url(r'^%s([a-zA-Z0-9\-_]+)/newthread/?$' % topic_prefix, views.createThread, name='createThread'),
    url(r'^%s(?P<topic_title>[a-zA-Z0-9\-_]+)/(?P<thread_id>\d+)/(?P<slug>[a-z0-9\-]+)?/?$' %
        topic_prefix, views.threadPage, name='threadPage'),
]



# Python standard imports
import json
import logging

# Core Dajngo imports
from django.http import JsonResponse, HttpResponse, Http404
from django.http import HttpResponseForbidden, HttpResponseBadRequest, HttpResponseRedirect
from django.shortcuts import render, redirect
from django.contrib.auth.models import User
from django.contrib.auth.decorators import user_passes_test
from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.views.decorators.http import require_POST

from django import VERSION as DJANGO_VERSION

if DJANGO_VERSION[:2] < (1, 10):
    from django.core.urlresolvers import reverse
else:
    from django.urls import reverse

# Third-party app imports
from ipware.ip import get_ip
from meta.views import Meta

# Imports from our apps
from djeddit.forms import TopicForm, ThreadForm, PostForm
from djeddit.models import Topic, Thread, Post, UserPostVote
from djeddit.templatetags.djeddit_tags import postScore
from djeddit.utils.utility_funcs import is_authenticated


# Create your views here.


def createThread(request, topic_title=None):
    if topic_title:
        try:
            if request.method == 'POST':
                topic = Topic.getTopic(topic_title)
                threadForm = ThreadForm(request.POST, prefix='thread')
                postForm = PostForm(request.POST, prefix='post')
                if threadForm.is_valid() and postForm.is_valid():
                    thread = threadForm.save(commit=False)
                    post = postForm.save(commit=False)
                    post.setMeta(request)
                    if is_authenticated(request):
                        post.created_by = request.user
                    post.save()
                    thread.op = post
                    thread.topic = topic
                    thread.save()
                    return HttpResponseRedirect(thread.relativeUrl)
            else:
                threadForm = ThreadForm(prefix='thread')
                postForm = PostForm(prefix='post')
            context = dict(threadForm=threadForm, postForm=postForm)
            return render(request, 'djeddit/create_thread.html', context)
        except Topic.DoesNotExist:
            pass
    return redirect('topics')


@require_POST
@user_passes_test(lambda u: u.is_superuser)
def deleteTopic(request, topic_title):
    try:
        topic = Topic.getTopic(topic_title)
    except Topic.DoesNotExist:
        raise Http404()
    topic.delete()
    return HttpResponse(json.dumps(dict(redirect=reverse('topics'))))


@require_POST
@user_passes_test(lambda u: u.is_superuser)
def lockThread(request, thread_id):
    try:
        thread = Thread.objects.get(id=thread_id)
    except Thread.DoesNotExist:
        raise Http404
    thread.locked = not thread.locked
    thread.save()
    return HttpResponse(json.dumps(dict(redirect=reverse('threadPage', args=[thread.topic.urlTitle, thread.pk]))))


@require_POST
@user_passes_test(lambda u: u.is_superuser)
def stickyThread(request, thread_id):
    try:
        thread = Thread.objects.get(id=thread_id)
    except Thread.DoesNotExist:
        raise Http404
    thread.is_stickied = not thread.is_stickied
    thread.save()
    return HttpResponse(json.dumps(dict(redirect=reverse('threadPage', args=[thread.topic.urlTitle, thread.pk]))))


def topicsPage(request):
    topics = Topic.objects.all()
    if request.method == 'POST':
        topicForm = TopicForm(request.POST)
        if topicForm.is_valid():
            topicForm.save()
            return redirect('topics')
        showForm = True
    else:
        topicForm = TopicForm()
        showForm = False
    context = dict(topics=topics, topicForm=topicForm, showForm=showForm)
    return render(request, 'djeddit/topics.html', context)


def topicPage(request, topic_title):
    try:
        topic = Topic.getTopic(topic_title)
    except Topic.DoesNotExist:
        raise Http404()
    # edit topic form
    if topic.urlTitle != topic_title:
        return redirect('topicPage', topic.urlTitle)
    if request.method == 'POST':
        if not request.user.is_superuser:
            return HttpResponseForbidden()
        form = TopicForm(request.POST, instance=topic)
        if form.is_valid():
            form.save()
            return redirect('topicPage', topic.urlTitle)
        showForm = True
    else:
        form = TopicForm(instance=topic)
        showForm = False
    threads = Thread.objects.filter(topic=topic).order_by('-is_stickied', '-op__created_on')
    context = dict(topic=topic, threads=threads, showCreatedBy=True, showTopic=False,
                   topicForm=form, showForm=showForm)
    return render(request, 'djeddit/topic.html', context)


def threadPage(request, topic_title='', thread_id='', slug=''):
    if topic_title and thread_id:
        try:
            topic = Topic.getTopic(topic_title)
            thread = Thread.objects.get(id=thread_id)
            if thread.topic == topic:
                if not slug or slug != thread.slug or topic.urlTitle != topic_title:
                    return HttpResponseRedirect(thread.relativeUrl)
                if thread.op.content:
                    description = thread.op.content[:160]
                else:
                    description = getattr(settings, "DJEDDIT_DESCRIPTION", "The link sharing and discussion portal")
                meta = Meta(
                    title=thread.title,
                    use_title_tag=True,
                    description=description,
                )
                thread.views += 1
                thread.save()
                context = dict(thread=thread, nodes=thread.op.getSortedReplies(), meta=meta)
                return render(request, 'djeddit/thread.html', context)
        except (Topic.DoesNotExist, Thread.DoesNotExist):
            pass
    raise Http404


def replyPost(request, post_uid=''):
    try:
        repliedPost = Post.objects.get(uid=post_uid)
        thread = repliedPost.thread
    except (Post.DoesNotExist, Thread.DoesNotExist):
        raise Http404
    if thread.locked:
        return HttpResponseForbidden()
    repliedUser = repliedPost.created_by.username if repliedPost.created_by else 'guest'
    if request.method == 'POST':
        postForm = PostForm(request.POST)
        if postForm.is_valid():
            post = postForm.save(commit=False)
            post.parent = repliedPost
            post.setMeta(request)

            if is_authenticated(request):
                post.created_by = request.user
            post.save()
            repliedPost.children.add(post)
        return HttpResponseRedirect(thread.relativeUrl)
    else:
        postForm = PostForm()
        postForm.fields['content'].label = ''
        context = dict(postForm=postForm, thread_id=thread.id, post_uid=post_uid, repliedUser=repliedUser)
        return render(request, 'djeddit/reply_form.html', context)


def editPost(request, post_uid=''):
    try:
        post = Post.objects.get(uid=post_uid)
        thread = post.thread
    except (Post.DoesNotExist, Thread.DoesNotExist):
        raise Http404
    if thread.locked or (request.user != post.created_by and not request.user.is_superuser):
        return HttpResponseForbidden()
    if request.method == 'POST':
        postForm = PostForm(request.POST, instance=post, prefix='post')
        threadForm = ThreadForm(request.POST, instance=thread, prefix='thread')
        if postForm.is_valid():
            postForm.save()
        if threadForm.is_valid():
            threadForm.save()
        return HttpResponseRedirect(thread.relativeUrl)
    else:
        postForm = PostForm(instance=post, prefix='post')
        if request.user.is_superuser and thread.op == post:
            threadForm = ThreadForm(instance=thread, prefix='thread')
        else:
            threadForm = None
            postForm.fields['content'].label = ''
        context = dict(postForm=postForm, threadForm=threadForm, post_uid=post.uid, thread=thread, post=post)
        return render(request, 'djeddit/edit_post.html', context)


@require_POST
@login_required
def votePost(request):
    try:
        post_uid = request.POST['post']
        vote_val = request.POST['vote']
        post = Post.objects.get(uid=post_uid)
    except (KeyError, ValueError, Post.DoesNotExist):
        return HttpResponseBadRequest()
    if post.created_by != request.user or request.user.is_superuser:
        try:
            userPostVote = UserPostVote.objects.get(user=request.user, post=post)
            oldval = userPostVote.val
            userPostVote.val = max(min(int(vote_val), 1), -1)
            userPostVote.save()
            voteDelta = userPostVote.val - oldval
        except UserPostVote.DoesNotExist:
            userPostVote = UserPostVote.objects.create(user=request.user, post=post, val=max(min(int(vote_val), 1), -1))
            voteDelta = userPostVote.val
        if voteDelta:
            if voteDelta > 0:
                if userPostVote.val:
                    post.upvotes += 1
                    if voteDelta == 2:
                        post.downvotes -= 1
                else:
                    post.downvotes -= 1
            else:
                if userPostVote.val:
                    post.downvotes += 1
                    if voteDelta == -2:
                        post.upvotes -= 1
                else:
                    post.upvotes -= 1
            post.save()
        scoreStr = postScore(post.score)
        return JsonResponse(dict(scoreStr=scoreStr, score=post.score))
    else:
        return HttpResponseForbidden()


@require_POST
@user_passes_test(lambda u: u.is_superuser)
def deletePost(request, post_uid):
    try:
        post = Post.objects.get(uid=post_uid)
    except Post.DoesNotExist:
        raise Http404
    thread = post.thread
    op = thread.op
    post_uid = post.uid
    post.delete()
    if op.uid == post_uid:
        url = reverse('topicPage', args=[thread.topic.urlTitle])
    else:
        url = thread.relativeUrl
    return HttpResponse(json.dumps(dict(redirect=url)))


def loadAdditionalReplies(request):
    """load additional replies to a given post."""
    try:
        post_uid = request.GET['post']
        # exclude all posts with these uids and their descendants
        excluded_uids = json.loads(request.GET['excluded'])
        post = Post.objects.get(uid=post_uid)
    except (KeyError, ValueError, json.decoder.JSONDecodeError, Post.DoesNotExist):
        return HttpResponseBadRequest()
    replies = post.getSortedReplies(excluded=excluded_uids)
    context = dict(thread=post.thread, nodes=replies)
    return render(request, 'djeddit/thread_recursetree.html', context)


def userSummary(request, username):
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        raise Http404
    threads = Thread.objects.filter(op__created_by=user)
    for t in threads:
        t.modified_on = t.op.modified_on
    replies = Post.objects.filter(created_by=user).exclude(parent=None)
    context = dict(items=sorted(list(threads) + list(replies), key=lambda n: n.modified_on, reverse=True),
                   tCount=threads.count(),
                   rCount=replies.count(),
                   tPoints=(sum(t.op.score for t in threads)),
                   rPoints=(sum(r.score for r in replies)),
                   pageUser=user)
    return render(request, 'djeddit/user_summary.html', context)


def userThreadsPage(request, username):
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        raise Http404
    created_threads = Thread.objects.filter(op__created_by=user)
    context = dict(threads=created_threads, showCreatedBy=False, showTopic=True, pageUser=user)
    return render(request, 'djeddit/user_threads.html', context)


def userRepliesPage(request, username):
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        raise Http404
    replies = Post.objects.filter(created_by=user, parent__isnull=False)
    context = dict(replies=replies, pageUser=user)
    return render(request, 'djeddit/user_replies.html', context)


@user_passes_test(lambda u: u.is_superuser)
def usersPage(request):
    """list users with their thread count and replies counts"""
    users = User.objects.all()
    for user in users:
        user.tCount = Thread.objects.filter(op__created_by=user).count()
        user.rCount = Post.objects.filter(created_by=user).exclude(parent=None).count()
    return render(request, 'djeddit/users_page.html', dict(Users=users))


@require_POST
@user_passes_test(lambda u: u.is_superuser)
def setUserStatus(request):
    """set user status to either active (is_active=True), banned(is_active=False), or admin(is_superuser=True)"""
    try:
        username = request.POST['username']
        status = request.POST['status']
    except KeyError:
        return HttpResponseBadRequest()
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        raise Http404()
    if status == 'active':
        user.is_active = True
        user.is_superuser = False
    elif status == 'banned':
        user.is_active = False
        user.is_superuser = False
    elif status == 'admin':
        user.is_active = True
        user.is_superuser = True
    else:
        return HttpResponseBadRequest()
    user.save()
    return HttpResponse()



from django.contrib.sitemaps import Sitemap

from .models import Thread


class ThreadSitemap(Sitemap):
    changefreq = "daily"
    priority = 0.5

    def items(self):
        return Thread.objects.all()

    def lastmod(self, obj):
        return obj.op.modified_on



# Core django imports
from django.conf import settings
from django import VERSION as DJANGO_VERSION
from django.utils.encoding import python_2_unicode_compatible
from django.core.validators import RegexValidator
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
if DJANGO_VERSION[:2] < (1, 10):
    from django.core.urlresolvers import reverse
else:
    from django.urls import reverse

# Third party imports
from mptt.models import MPTTModel, TreeForeignKey
from slugify import slugify
from ipware.ip import get_ip

# Our app imports
from djeddit.utils.utility_funcs import gen_uuid, wsi_confidence


class IntegerRangeField(models.IntegerField):
    def __init__(self, verbose_name=None, name=None, min_value=None, max_value=None, **kwargs):
        self.min_value, self.max_value = min_value, max_value
        models.IntegerField.__init__(self, verbose_name, name, **kwargs)

    def formfield(self, **kwargs):
        defaults = {'min_value': self.min_value, 'max_value': self.max_value}
        defaults.update(kwargs)
        return super(IntegerRangeField, self).formfield(**defaults)


class NamedModel(models.Model):
    class Meta:
        abstract = True

    def getName(self):
        return self.__class__.__name__


@python_2_unicode_compatible
class Topic(NamedModel):
    alphanumeric = RegexValidator(r'^[0-9a-zA-Z ]*$', 'Only alphanumeric characters are allowed.')
    title = models.CharField(max_length=20, blank=False, unique=True, validators=[alphanumeric])
    description = models.CharField(max_length=120, blank=True, default='')

    def __str__(self):
        return self.title

    def getThreadCount(self):
        return Thread.objects.filter(topic=self).count()

    @property
    def urlTitle(self):
        return self.title.replace(' ', '-')

    def get_absolute_url(self):
        return reverse('topicPage', args=[self.urlTitle])

    @staticmethod
    def getTopic(title):
        try:
            return Topic.objects.get(title=title)
        except ObjectDoesNotExist:
            try:
                return Topic.objects.get(title=title.replace('-', ' '))
            except ObjectDoesNotExist:
                return Topic.objects.get(title=title.replace('_', ' '))


@python_2_unicode_compatible
class Thread(NamedModel):
    title = models.CharField(max_length=70, blank=False)
    slug = models.SlugField(unique=False, null=True)
    url = models.URLField(max_length=120, blank=True, default='')
    views = models.IntegerField(blank=True, default=0)
    topic = models.ForeignKey(Topic, on_delete=models.CASCADE)
    op = models.ForeignKey('Post', related_name='+', on_delete=models.CASCADE)
    locked = models.BooleanField(blank=True, default=False)
    is_stickied = models.BooleanField(default=False)

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        self.slug = self._genSlug()
        super(Thread, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        try:
            self.op.delete()
        except Post.DoesNotExist:
            pass
        super(Thread, self).delete(*args, **kwargs)

    def _genSlug(self):
        slug = slugify(self.title, to_lower=True, max_length=80)
        return slug

    @property
    def relativeUrl(self):
        return reverse('threadPage', args=[self.topic.urlTitle, self.id, self.slug])

    def get_absolute_url(self):
        return self.relativeUrl


@python_2_unicode_compatible
class Post(MPTTModel, NamedModel):
    uid = models.UUIDField(max_length=8, primary_key=True, default=gen_uuid, editable=False)
    content = models.TextField(blank=True, default='')
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.CASCADE, related_name='+')
    created_on = models.DateTimeField(auto_now_add=True, auto_now=False)
    modified_on = models.DateTimeField(auto_now_add=False, auto_now=True)
    parent = TreeForeignKey('self', null=True, blank=True, related_name='children', db_index=True, on_delete=models.CASCADE)
    _upvotes = models.IntegerField(blank=True, default=0)
    _downvotes = models.IntegerField(blank=True, default=0)
    wsi = models.FloatField(blank=True, default=0) # Wilson score interval
    ip_address = models.GenericIPAddressField(blank=True, null=True)
    user_agent = models.CharField(max_length=150, blank=True, null=True)

    def __init__(self, *args, **kwargs):
        super(Post, self).__init__(*args, **kwargs)
        Post.upvotes = property(lambda self: self._upvotes, Post._voteSetterWrapper('_upvotes'))
        Post.downvotes = property(lambda self: self._downvotes, Post._voteSetterWrapper('_downvotes'))

    class MPTTMetta:
        order_insertion_by = ['created_on']

    def __str__(self):
        return self.content[:70]

    @staticmethod
    def _voteSetterWrapper(attr):
        def voteSetter(self, value):
            setattr(self, attr, max(0, value))
            self.wsi = wsi_confidence(self._upvotes, self._downvotes)
        return voteSetter

    @property
    def thread(self): # TODO: thread should be stored in Post
        post = self
        while post.parent:
            post = post.parent
        return Thread.objects.get(op=post)

    @property
    def score(self):
        return self.upvotes - self.downvotes

    def getReplies(self, excluded=()):
        """:param excluded: exclude all posts with these uids and their descendants"""
        replies = Post.objects.filter(parent=self.uid).exclude(uid__in=excluded)
        for reply in replies:
            replies |= reply.getReplies(excluded=excluded)
        return replies

    def getSortedReplies(self, limit=50, by_wsi=True, excluded=()):
        """
        :param limit: number of replies to return
        :param by_wsi: sort replies by wsi score or by creation date
        :param excluded: uids or excluded replies
        """
        excluded = list(excluded)
        order_field = '-wsi' if by_wsi else 'created_on'
        replies = list(self.getReplies(excluded=excluded).order_by(order_field)[:limit])
        self._getPostsWithChildren(replies)
        sorted_replies = []
        for p in replies:
            sorted_replies.append(p)
            sorted_replies += p.getChildrenList()
        return sorted_replies

    def _getPostsWithChildren(self, replies):
        for p in list(replies):
            if not hasattr(p, 'included_children'):
                p.included_children = []
            if p.parent != self:
                if p.parent not in replies:
                    # add missing parents
                    current_p = p
                    while True:
                        current_p.parent._addToIncludedChildren(current_p)
                        current_p = current_p.parent
                        if current_p in (replies + [self]):
                            break
                    if current_p in replies:
                        replies[replies.index(current_p)] = current_p
                else:
                    p.parent = replies[replies.index(p.parent)]
                    p.parent._addToIncludedChildren(p)
                replies.remove(p)

    def _addToIncludedChildren(self, post):
        if not hasattr(self, 'included_children'):
            self.included_children = [post]
        else:
            self.included_children.append(post)

    def getChildrenList(self):
        children = []
        for p in self.included_children:
            children.append(p)
            if p.included_children:
                children += p.getChildrenList()
        return children

    def setMeta(self, request):
        """update post ip_address & user_agent attributes"""
        ip = get_ip(request)
        if ip is not None:
            self.ip_address = ip
        ua = request.META.get('HTTP_USER_AGENT', '')
        if ua:
            self.user_agent = ua


class UserPostVote(NamedModel):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='+')
    post = models.ForeignKey(Post, related_name='+', on_delete=models.CASCADE)
    val = IntegerRangeField(blank=True, default=0, min_value=-1, max_value=1)



import json

from six.moves.urllib.parse import urlparse

from django.contrib.auth.models import User
from django.conf import settings


def createUser(username='user', email='user@example.com', password='pass', **kwargs):
    user = User(username=username, email=email, **kwargs)
    user.set_password(password)
    user.save()
    return user


class TestCalls(object):
    def __init__(self, template=''):
        self.template = template
        self.user = None

    def login(self, username='', password=''):
        username = username or self.username
        password = password or self.password
        if username and password:
            self.client.login(username=username, password=password)

    @classmethod
    def _setup_user(cls, username, email, password, **kwargs):
        cls.user = createUser(username, email, password, **kwargs)
        cls.username = username
        cls.password = password

    def _create_user_and_login(self, username='', email='', password='', create_admin=False):
        if create_admin:
            username = username or 'admin'
            email = email or 'admin@example.com'
        else:
            username = username or 'user'
            email = email or 'user@example.com'
        password = password or 'pass'
        user = createUser(username, email, password, is_superuser=create_admin)
        self.login(username, password)
        return user

    def _test_call_view_loads(self, url, data=None):
        data = data or {}
        response = self.client.get(url, data)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(self.template)

    def _test_call_view_submit(self, url, code=200, data=None):
        data = data or {}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, code)
        if self.template:
            self.assertTemplateUsed(self.template)

    def _test_call_view_redirected_login(self, url, data=None):
        data = data or {}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(urlparse(response.url).path, settings.LOGIN_URL)

    def _test_call_view_redirects(self, url, redirected_url, data=None):
        data = data or {}
        response = self.client.post(url, data, follow=True)
        self._assert_redirects(response, redirected_url)

    def _test_call_view_code(self, url, code, data=None, post=False):
        data = data or {}
        response = self.client.post(url, data) if post else self.client.get(url, data)
        self.assertEqual(response.status_code, code)

    def _assert_redirects(self, response, redirected_url):
        if response.status_code in (200, 302):
            try:
                self.assertRedirects(response, redirected_url)
            except AssertionError:
                content = json.loads(response.content.decode('utf-8'))
                self.assertEqual(content['redirect'], redirected_url)
        else:
            raise ValueError('invalid status code %s' % response.status_code)

